//
//  search.cpp
//  fornax3
//
//  Created by Anders on 28/04/2019.
//
#include <thread>
#include <inttypes.h>
#include "search.hpp"
#include "../tt/transpositions.hpp"
#include "../moving/movegen.hpp"
#include "../moving/movepick.hpp"
#include "../moving/makemove.hpp"
#include "../parsing.hpp"
#include "../libs/stopwatch.h"
#include "../libs/iolog.hpp"
#include "../eval_defs.h"
#include "../supplementary/time_management.hpp"

#define CURRMOVE_NODE_THRESHOLD 2500000
#define SEARCH_DEPTH_MAX 128

static std::atomic<bool> async_is_currently_searching (false);
static std::atomic<bool> async_is_stop_command_given (false);
static std::atomic<uint64_t> node_count(0);

#ifdef DEBUGLOG
static uint64_t betacuts = 0;
#define DEBUG_STATS_INCREMENT_BETA_CUTS betacuts++;
static uint64_t quiesceCount = 0;
#define DEBUG_STATS_INCREMENT_QS_COUNT quiesceCount++;
static uint64_t transpositionHits = 0;
#define DEBUG_STATS_INCREMENT_TT_HITS transpositionHits++;
static uint64_t transpositionCuts = 0;
#define DEBUG_STATS_INCREMENT_TT_CUTS transpositionCuts++;
static uint64_t nullcuts = 0;
#define DEBUG_STATS_INCREMENT_NULL_CUTS nullcuts++;
#define RESET_STATS node_count = 1;\
betacuts = 0;\
quiesceCount = 0;\
transpositionHits = 0;\
transpositionCuts = 0;\
nullcuts = 0

#else
#define DEBUG_STATS_INCREMENT_BETA_CUTS
#define DEBUG_STATS_INCREMENT_QS_COUNT
#define DEBUG_STATS_INCREMENT_TT_HITS
#define DEBUG_STATS_INCREMENT_TT_CUTS
#define DEBUG_STATS_INCREMENT_NULL_CUTS
#define RESET_STATS node_count = 1
#endif

typedef struct {
  move bestmove;
  move ponder;
  eval score;
  long duration_ms;
} SearchResult;


static inline uint64_t getnps(long duration) {
  long time = duration | 1;
  return ((node_count | 1) / (uint32_t) time) * 1000;
}

static inline void printline(Board* board, uint8_t depth, eval eval, move move, long duration) {
  uint64_t nodes = node_count;
  
  IO_PRINT("info depth %d score cp %d nodes %" PRIu64 " nps %" PRIu64, depth, eval, nodes, getnps(duration));
#ifdef DEBUGLOG
  IO_PRINT(" tthits %" PRIu64, transpositionHits);
  IO_PRINT(" ttcuts %" PRIu64, transpositionCuts);
  IO_PRINT(" qnodes %" PRIu64, quiesceCount);
  IO_PRINT(" betacuts %" PRIu64, betacuts);
  IO_PRINT(" nullcuts %" PRIu64, nullcuts);
#endif
  IO_PRINT(" pv ");
  parsing_printpv(board, move, depth);
  IO_PRINT("\n");
  IO_FLUSH();
}

static inline eval probe_tt(ttkey hash, move* ttmove, int depth, eval alpha, eval beta) {
  Bucket* b = transpositions_get(hash);
  if (b->hash == hash) {
    *ttmove = b->ttmove;
    
    DEBUG_STATS_INCREMENT_TT_HITS;
    
    if (b->depth >= depth) {
      DEBUG_STATS_INCREMENT_TT_CUTS;
      tttype type = b->type;
      if(type == TT_TYPE_EXACT) return b->tteval;
      if(type == TT_TYPE_ALPHA && b->tteval <= alpha) return alpha;
      if(type == TT_TYPE_BETA && b->tteval >= beta) return beta;
    }
  }
  
  return EVAL_NONE;
}


template <color active>
static constexpr bool ischeck(const Board* node) {
  constexpr color opp = COLOR_OPPOSITE(active);
  return node->side[active].pieces[KING] & node->side[opp].attackmask;
}

#define REPETITION_HISTORY_SIZE 18
static inline bool is_repetition(const Board* board) {
  if (board->halfmovecount >= 100) return true;
  ttkey hash = board_get_currinfo(board)->hash;
  
  uint8_t end = board->ply - REPETITION_HISTORY_SIZE;
  
  for (uint8_t i = board->ply - 2; i != end; i -= 2) {
    ttkey candidate = board->plyinfo[i].hash;
    
    if (hash == candidate) {
      return true;
    }
  }
  return false;
}


template <color active>
static eval quiesce(Board* node, uint8_t depth, eval alpha, eval beta) {
  constexpr color opp = COLOR_OPPOSITE(active);
  
  DEBUG_STATS_INCREMENT_QS_COUNT;
  node_count++;
  
  bool check = ischeck<active>(node);
  
  if (check) {
    depth++;
  }
  else {
    eval pat = eval_evaluate(node);
    if (depth == 0) return pat;
    if (pat >= beta) return beta;
    if (pat > alpha) alpha = pat;
  }
  
  evalmove moves[128];
  uint8_t size = movegen_generate<active, false>(node, moves);
  
  if (size == 0) {
    eval e = check ? EVAL_CHECKMATE + 1 : alpha;
    return e;
  }
  
  movepick_annotate<active>(node, moves, size, 0);
  for (uint8_t n = 0; n < size; ++n) {
    evalmove m = movepick_find_best(moves, size, n);
    
    if (!check && MOVE_VALUE(m) < EVAL_QUIESCE_LIMIT_VALUE) break;
    
    makemove_make<active>(node, (move) m);
    eval eval = -quiesce<opp>(node, depth - 1, -beta, -alpha);
    makemove_takeback<opp>(node, (move) m);
    if (eval >= beta) {
      DEBUG_STATS_INCREMENT_BETA_CUTS;
      return beta;
    }
    if (eval > alpha) {
      alpha = eval;
    }
  }
  return alpha;
}

template <color active>
static uint8_t late_move_reduction(uint8_t depth,
                                   move move,
                                   piecetype piece,
                                   piecetype capture,
                                   bool is_check,
                                   bool is_checkingmove) {
  
  if (depth < 2
      || is_check
      || is_checkingmove
      || capture != NONE) return 0;
  assert(piece != NONE);
  square y = SQUARE_Y(MOVE_GET_ORIGIN(move));
  bool is_danger_pawn = piece == PAWN && (active ? y < 5 : y > 2);
  
  if (is_danger_pawn) return 0;
  return 1;
}

template <color active, bool is_pv>
static eval search_node(Board* node, uint8_t depth, eval alpha, eval beta);

template <color active>
static inline eval null_search(Board* node, uint8_t depth, eval beta) {
  constexpr color opp = COLOR_OPPOSITE(active);
  makemove_make_null<active>(node);
  uint8_t reduced = depth - 2;
  reduced -= depth / 5;
  eval eval = -search_node<opp, false>(node, reduced - 1, -beta, -beta + 1);
  makemove_takeback_null<opp>(node);
  return eval;
}

template <color active, bool do_null>
static eval search_node(Board* node, uint8_t depth, eval alpha, eval beta) {
  constexpr color opp = COLOR_OPPOSITE(active);
  if (depth == 0) {
    return quiesce<active>(node, 9, alpha, beta);
  }
  assert(depth > 0);
  node_count++;
  
  if (is_repetition(node)) return eval_calculate_contempt(node);
  Plyinfo *info = &node->plyinfo[node->ply];
  
  move ttmove = 0;
  eval tteval;
  if ((tteval = probe_tt(info->hash, &ttmove, depth, alpha, beta)) != EVAL_NONE) {
    return tteval;
  }
  
  bool check = ischeck<active>(node);
  
  evalmove moves[128];
  uint8_t size = movegen_generate<active>(node, moves);
  
  if (size == 0) {
    eval e = check ? (EVAL_CHECKMATE - depth) : eval_calculate_contempt(node);
    transpositions_put(info->hash, depth, e, 0, TT_TYPE_EXACT);
    return e;
  }
  else if (size == 1) {
    makemove_make<active>(node, (move) moves[0]);
    eval e = -search_node<opp, do_null>(node, depth, -beta, -alpha);
    makemove_takeback<opp>(node, (move) moves[0]);
    return e;
  }
  
  if (check) {
    if (node->side[opp].eval_material > EVAL_MATERIAL_VALUE_EXTEND_CHECK_THRESHOLD) {
      depth++;
    }
  }
  else if (depth <= 2 && beta > EVAL_CHECKMATE) {
    eval threshold = depth == 1 ? EVAL_PIECE_BISHOPS : EVAL_PIECE_ROOKS;
    eval eval = eval_evaluate(node) - threshold;
    if (eval > beta) {
      DEBUG_STATS_INCREMENT_NULL_CUTS;
      return eval;
    }
  }
  else if (do_null && depth > 2 && node->side[active].eval_material > EVAL_MATERIAL_VALUE_NULL_THRESHOLD) {
    eval eval = null_search<active>(node, depth, beta);
    if(eval >= beta) {
      DEBUG_STATS_INCREMENT_NULL_CUTS;
      return beta;
    }
  }
  
  movepick_annotate<active>(node, moves, size, ttmove);
  
  move m = (move) movepick_find_best(moves, size, 0);
  tttype tttype = TT_TYPE_ALPHA;
  
  /*PVS: search pv move with full beta range*/
  makemove_make<active>(node, m);
  eval criticalEval = -search_node<opp, false>(node, depth - 1, -beta, -alpha);
  makemove_takeback<opp>(node, m);
  
  if (criticalEval >= beta) {
    DEBUG_STATS_INCREMENT_BETA_CUTS;
    transpositions_put(info->hash, depth, criticalEval, m, TT_TYPE_BETA);
    movepick_set_killer(node->ply, m, ttmove);
    return criticalEval;
  }
  if (criticalEval > alpha) {
    alpha = criticalEval;
    tttype = TT_TYPE_EXACT;
  }
  
  move criticalMove = m;
  
  for (uint8_t n = 1; n < size; ++n) {
    if (async_is_stop_command_given) return 0;
    m = (move) movepick_find_best(moves, size, n);
    piecetype piece = node->piecemap[MOVE_GET_ORIGIN(m)];
    piecetype capture = node->piecemap[MOVE_GET_DEST(m)];
    
    makemove_make<active>(node, m);
    
    uint8_t reduction = late_move_reduction<active>(depth, m, piece, capture, check, ischeck<opp>(node));
    eval eval = -search_node<opp, true>(node, depth - reduction - 1, -alpha - 1, -alpha);
    
    /*PVS: if outside bounds search again with full range*/
    if (eval > alpha && (eval < beta || reduction > 0)) {
      makemove_genmaps(node);
      eval = -search_node<opp, false>(node, depth - 1, -beta, -alpha);
    }
    makemove_takeback<opp>(node, m);
    if (eval > criticalEval) {
      criticalEval = eval;
      criticalMove = m;
    }
    if (eval >= beta) {
      DEBUG_STATS_INCREMENT_BETA_CUTS;
      transpositions_put(info->hash, depth, eval, criticalMove, TT_TYPE_BETA);
      movepick_set_killer(node->ply, m, ttmove);
      return eval;
    }
    if (eval > alpha) {
      alpha = eval;
      tttype = TT_TYPE_EXACT;
    }
  }
  transpositions_put(info->hash, depth, criticalEval, criticalMove, tttype);
  return criticalEval;
}

template <color active>
static inline move search_find_ponder(Board* node, move move) {
  constexpr color opp = COLOR_OPPOSITE(active);
  makemove_make<active>(node, move);
  ttkey hash = node->plyinfo[node->ply].hash;
  Bucket* bucket = transpositions_get(hash);
  makemove_takeback<opp>(node, move);
  makemove_genmaps(node);
  if (bucket->hash != hash) return MOVE_NONE;
  return bucket->ttmove;
}

template <color active, bool print>
static inline void search_set_results(SearchResult* result, Board* root, uint8_t depth, eval eval, move move, long starttime) {
  long duration = stopwatch_wall_stop(starttime);
  if (print) printline(root, depth, eval, move, duration);
  result->bestmove = move;
  result->ponder = search_find_ponder<active>(root, move);
  result->score = eval;
  result->duration_ms = duration;
}

template <color active, bool print>
static eval search_root(SearchResult* result,
                        const evalmove rootmoves[128],
                        uint8_t rootmovesize,
                        Board* node,
                        uint8_t depth,
                        bool printcurr) {
  constexpr color opp = COLOR_OPPOSITE(active);
  RESET_STATS;
  
  long starttime = stopwatch_wall_start();
  Plyinfo *info = &node->plyinfo[node->ply];
  
  bool check = ischeck<active>(node);
  
  if (rootmovesize == 0) {
    eval e = check ? (EVAL_CHECKMATE) : 0;
    search_set_results<active, print>(result, node, depth, e, 0, 0);
    return 0;
  }
  
  move ttmove = 0;
  eval tteval;
  if ((tteval = probe_tt(info->hash, &ttmove, depth, EVAL_ALPHASTART, EVAL_BETASTART)) != EVAL_NONE) {
    if (!is_repetition(node)) {
      search_set_results<active, print>(result, node, depth, tteval, ttmove, starttime);
      return tteval;
    }
  }
  node->active_initial = active;
  
  uint8_t currmovenumber = 0;
  
  evalmove moves[128];
  for (int i = 0; i < rootmovesize; ++i) moves[i] = rootmoves[i];
  movepick_annotate_ttmove(moves, rootmovesize, ttmove);
  
  move m = (move) movepick_find_best(moves, rootmovesize, 0);
  
  /*PVS: search pv move with full beta range*/
  makemove_make<active>(node, m);
  eval alpha = -search_node<opp, false>(node, depth - 1, EVAL_ALPHASTART, EVAL_BETASTART);
  makemove_takeback<opp>(node, m);
  
  move criticalMove = m;
  
  for (uint8_t n = 1; n < rootmovesize; ++n) {
    if (async_is_stop_command_given) return 0;
    
    m = (move) movepick_find_best(moves, rootmovesize, n);
    piecetype piece = node->piecemap[MOVE_GET_ORIGIN(m)];
    piecetype capture = node->piecemap[MOVE_GET_DEST(m)];
    
    currmovenumber++;
    if (printcurr && print) {
      parsing_printcurrmove(m, depth, currmovenumber);
    }
    
    makemove_make<active>(node, m);
    uint8_t reduction = late_move_reduction<active>(depth, m, piece, capture, check, ischeck<opp>(node));
    eval eval = -search_node<opp, true>(node, depth - reduction - 1, -alpha - 1, -alpha);
    
    /*PVS: if outside bounds search again with full range*/
    if (eval > alpha) {
      makemove_genmaps(node);
      eval = -search_node<opp, true>(node, depth - 1, EVAL_ALPHASTART, -alpha);
    }
    makemove_takeback<opp>(node, m);

    if (eval > alpha) {
      alpha = eval;
      criticalMove = m;
    }
  }
  transpositions_put(info->hash, depth, alpha, criticalMove, TT_TYPE_EXACT);
  
  if (!async_is_stop_command_given) {
    search_set_results<active, print>(result, node, depth, alpha, criticalMove, starttime);
  }
  
  makemove_genmaps(node);
  return alpha;
}

/* at depth 1 all moves are annotated with quiesce results for better sorting*/
template <color active, bool print>
static uint8_t search_depth_1(SearchResult* result, Board* node, evalmove moves[128]) {
  constexpr color opp = COLOR_OPPOSITE(active);
  RESET_STATS;
  
  uint8_t size = movegen_generate<active>(node, moves);
  if (size == 0) {
    eval e = ischeck<active>(node) ? (EVAL_CHECKMATE) : 0;
    search_set_results<active, print>(result, node, 1, e, 0, 0);
    return 0;
  }
  
  eval criticalEval = EVAL_NONE;
  move criticalMove = MOVE_NONE;
  
  for (uint8_t n = 0; n < size; n++) {
    move m = (move) moves[n];
    makemove_make<active>(node, m);
    eval eval = -quiesce<opp>(node, 9, EVAL_ALPHASTART, EVAL_BETASTART);
    if (eval > criticalEval) {
      criticalEval = eval;
      criticalMove = m;
    }
    
    makemove_takeback<opp>(node, m);
    moves[n] |= (unsigned int) (eval << 16);
  }
  search_set_results<active, print>(result, node, 1, criticalEval, criticalMove, 0);
  
  makemove_genmaps(node);
  return size;
}

template <color active, bool print>
static move search_iterative_templ(const Board* board, uint8_t depth) {
#ifdef DEBUGLOG
  long start = stopwatch_wall_start();
#endif
  bool printcurr = false;
  
  SearchResult result = {0, 0, EVAL_NONE, 0};
  
  Board node = *board; // copy board
  evalmove rootmoves[128];
  uint8_t rootmovesize = search_depth_1<active, print>(&result, &node, rootmoves);

  uint8_t currdepth = 2;
  while (currdepth <= depth) {
    search_root<active, print>(&result, rootmoves, rootmovesize, &node, currdepth, printcurr);
    
    currdepth++;
    printcurr = node_count > CURRMOVE_NODE_THRESHOLD;
    
    if (currdepth == 0) break;
    if (async_is_stop_command_given) break;
    if (time_management_check_stop()) break;
  }
  time_management_send_stopped();
  
#ifdef DEBUGLOG
  if (print) IO_PRINT("info string - searched for %.3fs\n", (double) stopwatch_wall_stop(start) / 1000);
#endif
  
  async_is_currently_searching = false;
  async_is_stop_command_given = false;
  if (print) parsing_printbestmove(result.bestmove, result.ponder);
  
  return result.bestmove;
}

move search_iterative(const Board* board, uint8_t depth, bool print) {
  return board->active
  ? print ? search_iterative_templ<BLACK, true>(board, depth) : search_iterative_templ<BLACK, false>(board, depth)
  : print ? search_iterative_templ<WHITE, true>(board, depth) : search_iterative_templ<WHITE, false>(board, depth);
}

std::future<move> search_async_depth(const Board* board, uint8_t depth, bool print) {
#ifdef DEBUGLOG
  if (async_is_currently_searching) IO_PRINT("info string - warn: already in a search!\n");
#endif
  
  if (async_is_currently_searching) {
    return move_none();
  }
  async_is_currently_searching = true;
  async_is_stop_command_given = false;
  return std::async (std::launch::async, search_iterative, board, depth, print);
}

std::future<move> search_async_infinite(const Board* board, bool print) {
  return search_async_depth(board, SEARCH_DEPTH_MAX, print);
}

std::future<move> search_async_time(const Board* board, long ms, bool print) {
  std::thread(stopwatch_sleep, ms, search_async_stop).detach();
  return search_async_depth(board, SEARCH_DEPTH_MAX, print);
}

void search_async_stop(void) {
#ifdef DEBUGLOG
  IO_PRINT("info string - stopping ...\n");
#endif
  async_is_stop_command_given = true;
}

void search_async_printinfo(void) {
  uint64_t nodes = node_count;
  IO_PRINT("info nodes %" PRIu64 "\n", nodes);
}
